@@ -5,6 +5,7 @@
 #include <irc/irc_client.hpp>
 #include <bridge/bridge.hpp>
 #include <irc/irc_user.hpp>
+#include <utils/base64.hpp>
 
 #include <logger/logger.hpp>
 #include <config/config.hpp>
@@ -81,7 +82,18 @@ static const std::unordered_map<std::string,
   {"PONG", {&IrcClient::on_pong, {0, 0}}},
   {"KICK", {&IrcClient::on_kick, {3, 0}}},
   {"INVITE", {&IrcClient::on_invite, {2, 0}}},
-
+  {"CAP", {&IrcClient::on_cap, {3, 0}}},
+#ifdef WITH_SASL
+  {"AUTHENTICATE", {&IrcClient::on_authenticate, {1, 0}}},
+  {"900", {&IrcClient::on_sasl_login, {3, 0}}},
+  {"902", {&IrcClient::on_sasl_failure, {2, 0}}},
+  {"903", {&IrcClient::on_sasl_success, {0, 0}}},
+  {"904", {&IrcClient::on_sasl_failure, {2, 0}}},
+  {"905", {&IrcClient::on_sasl_failure, {2, 0}}},
+  {"906", {&IrcClient::on_sasl_failure, {2, 0}}},
+  {"907", {&IrcClient::on_sasl_failure, {2, 0}}},
+  {"908", {&IrcClient::on_sasl_failure, {2, 0}}},
+#endif
   {"401", {&IrcClient::on_generic_error, {2, 0}}},
   {"402", {&IrcClient::on_generic_error, {2, 0}}},
   {"403", {&IrcClient::on_generic_error, {2, 0}}},
@@ -135,7 +147,7 @@ IrcClient::IrcClient(std::shared_ptr<Poller>& poller, std::string hostname,
                      std::string realname, std::string user_hostname,
                      Bridge& bridge):
   TCPClientSocketHandler(poller),
-  hostname(std::move(hostname)),
+  hostname(hostname),
   user_hostname(std::move(user_hostname)),
   username(std::move(username)),
   realname(std::move(realname)),
@@ -143,7 +155,14 @@ IrcClient::IrcClient(std::shared_ptr<Poller>& poller, std::string hostname,
   bridge(bridge),
   welcomed(false),
   chanmodes({"", "", "", ""}),
-  chantypes({'#', '&'})
+  chantypes({'#', '&'}),
+  tokens_bucket(this->get_throttle_limit(), 1s, [this]() {
+    if (message_queue.empty())
+      return true;
+    this->actual_send(std::move(this->message_queue.front()));
+    this->message_queue.pop_front();
+    return false;
+  }, "TokensBucket" + this->hostname + this->bridge.get_jid())
 {
 #ifdef USE_DATABASE
   auto options = Database::get_irc_server_options(this->bridge.get_bare_jid(),
@@ -171,6 +190,7 @@ IrcClient::~IrcClient()
   // This event may or may not exist (if we never got connected, it
   // doesn't), but it's ok
   TimedEventsManager::instance().cancel("PING" + this->hostname + this->bridge.get_jid());
+  TimedEventsManager::instance().cancel("TokensBucket" + this->hostname + this->bridge.get_jid());
 }
 
 void IrcClient::start()
@@ -224,6 +244,7 @@ void IrcClient::on_connection_failed(const std::string& reason)
                                             "cancel", "item-not-found",
                                             "", reason);
         }
+      this->channels_to_join.clear();
     }
   else                          // try the next port
     this->start();
@@ -264,18 +285,45 @@ void IrcClient::on_connected()
         }
     }
 
-  this->send_message({"CAP", {"REQ", "multi-prefix"}});
-  this->send_message({"CAP", {"END"}});
+  this->send_gateway_message("Connected to IRC server"s + (this->use_tls ? " (encrypted)": "") + ".");
+
+  this->capabilities["multi-prefix"] = {[]{}, []{}};
 
 #ifdef USE_DATABASE
   auto options = Database::get_irc_server_options(this->bridge.get_bare_jid(),
                                                   this->get_hostname());
-  if (!options.col<Database::Pass>().empty())
+
+  const auto& server_password = options.col<Database::Pass>();
+
+  if (!server_password.empty())
     this->send_pass_command(options.col<Database::Pass>());
 #endif
 
-  this->send_nick_command(this->current_nick);
+#ifdef WITH_SASL
+  const auto& sasl_password = options.col<Database::SaslPassword>();
+  if (!sasl_password.empty())
+    {
+      this->capabilities["sasl"] = {
+          [this]
+          {
+            this->send_message({"AUTHENTICATE", {"PLAIN"}});
+            log_warning("negociating SASL now...");
+          },
+          []
+          {
+            log_warning("SASL not supported by the server, disconnecting.");
+          }
+      };
+      this->sasl_state = SaslState::needed;
+    }
+#endif
 
+  {
+    for (const auto &pair : this->capabilities)
+      this->send_message({ "CAP", {"REQ", pair.first}});
+  }
+
+  this->send_nick_command(this->current_nick);
 #ifdef USE_DATABASE
   if (Config::get("realname_customization", "true") == "true")
     {
@@ -286,13 +334,8 @@ void IrcClient::on_connected()
       this->send_user_command(username, realname);
     }
   else
-    this->send_user_command(this->username, this->realname);
-#else
-  this->send_user_command(this->username, this->realname);
 #endif
-  this->send_gateway_message("Connected to IRC server"s + (this->use_tls ? " (encrypted)": "") + ".");
-  this->send_pending_data();
-  this->bridge.on_irc_client_connected(this->get_hostname());
+  this->send_user_command(this->username, this->realname);
 }
 
 void IrcClient::on_connection_close(const std::string& error_msg)
@@ -317,9 +360,21 @@ IrcChannel* IrcClient::get_channel(const std::string& n)
     }
   catch (const std::out_of_range& exception)
     {
-      this->channels.emplace(name, std::make_unique<IrcChannel>());
+      return this->channels.emplace(name, std::make_unique<IrcChannel>()).first->second.get();
+    }
+}
+
+const IrcChannel* IrcClient::find_channel(const std::string& n) const
+{
+  const std::string name = utils::tolower(n);
+  try
+    {
       return this->channels.at(name).get();
     }
+  catch (const std::out_of_range& exception)
+    {
+      return nullptr;
+    }
 }
 
 bool IrcClient::is_channel_joined(const std::string& name)
@@ -351,12 +406,14 @@ void IrcClient::parse_in_buffer(const size_t)
         {
           const auto& limits = it->second.second;
           // Check that the Message is well formed before actually calling
-          // the callback. limits.first is the min number of arguments,
-          // second is the max
-          if (message.arguments.size() < limits.first ||
-              (limits.second > 0 && message.arguments.size() > limits.second))
+          // the callback.
+          const auto args_size = message.arguments.size();
+          const auto min = limits.first;
+          const auto max = limits.second;
+          if (args_size < min ||
+              (max > 0 && args_size > max))
             log_warning("Invalid number of arguments for IRC command “", message.command,
-                        "”: ", message.arguments.size());
+                        "”: ", args_size);
           else
             {
               const auto& cb = it->second.first;
@@ -378,25 +435,39 @@ void IrcClient::parse_in_buffer(const size_t)
     }
 }
 
-void IrcClient::send_message(IrcMessage&& message)
+void IrcClient::actual_send(std::pair<IrcMessage, MessageCallback>&& message_pair)
 {
-  log_debug("IRC SENDING: (", this->get_hostname(), ") ", message);
-  std::string res;
-  if (!message.prefix.empty())
-    res += ":" + std::move(message.prefix) + " ";
-  res += message.command;
-  for (const std::string& arg: message.arguments)
-    {
-      if (arg.find(' ') != std::string::npos ||
-          (!arg.empty() && arg[0] == ':'))
-        {
-          res += " :" + arg;
-          break;
-        }
-      res += " " + arg;
-    }
-  res += "\r\n";
-  this->send_data(std::move(res));
+  const IrcMessage& message = message_pair.first;
+  const MessageCallback& callback = message_pair.second;
+   log_debug("IRC SENDING: (", this->get_hostname(), ") ", message);
+    std::string res;
+    if (!message.prefix.empty())
+      res += ":" + message.prefix + " ";
+    res += message.command;
+    for (const std::string& arg: message.arguments)
+      {
+        if (arg.find(' ') != std::string::npos
+            || (!arg.empty() && arg[0] == ':'))
+          {
+            res += " :" + arg;
+            break;
+          }
+        res += " " + arg;
+      }
+    res += "\r\n";
+    this->send_data(std::move(res));
+
+    if (callback)
+      callback(this, message);
+ }
+
+void IrcClient::send_message(IrcMessage message, MessageCallback callback, bool throttle)
+{
+  auto message_pair = std::make_pair(std::move(message), std::move(callback));
+  if (this->tokens_bucket.use_token() || !throttle)
+    this->actual_send(std::move(message_pair));
+  else
+    message_queue.push_back(std::move(message_pair));
 }
 
 void IrcClient::send_raw(const std::string& txt)
@@ -447,12 +518,12 @@ void IrcClient::send_topic_command(const std::string& chan_name, const std::stri
 
 void IrcClient::send_quit_command(const std::string& reason)
 {
-  this->send_message(IrcMessage("QUIT", {reason}));
+  this->send_message(IrcMessage("QUIT", {reason}), {}, false);
 }
 
 void IrcClient::send_join_command(const std::string& chan_name, const std::string& password)
 {
-  if (this->welcomed == false)
+  if (!this->welcomed)
     {
       const auto it = std::find_if(begin(this->channels_to_join), end(this->channels_to_join),
                                    [&chan_name](const auto& pair) { return std::get<0>(pair) == chan_name; });
@@ -466,10 +537,11 @@ void IrcClient::send_join_command(const std::string& chan_name, const std::strin
   this->start();
 }
 
-bool IrcClient::send_channel_message(const std::string& chan_name, const std::string& body)
+bool IrcClient::send_channel_message(const std::string& chan_name, const std::string& body,
+                                     MessageCallback callback)
 {
   IrcChannel* channel = this->get_channel(chan_name);
-  if (channel->joined == false)
+  if (!channel->joined)
     {
       log_warning("Cannot send message to channel ", chan_name, ", it is not joined");
       return false;
@@ -489,7 +561,7 @@ bool IrcClient::send_channel_message(const std::string& chan_name, const std::st
                          ::strlen(":!@ PRIVMSG ") - chan_name.length() - ::strlen(" :\r\n");
   const auto lines = cut(body, line_size);
   for (const auto& line: lines)
-    this->send_message(IrcMessage("PRIVMSG", {chan_name, line}));
+    this->send_message(IrcMessage("PRIVMSG", {chan_name, line}), callback);
   return true;
 }
 
@@ -1123,8 +1195,6 @@ void IrcClient::on_channel_bad_key(const IrcMessage& message)
 
 void IrcClient::on_channel_mode(const IrcMessage& message)
 {
-  // For now, just transmit the modes so the user can know what happens
-  // TODO, actually interprete the mode.
   Iid iid;
   iid.set_local(message.arguments[0]);
   iid.set_server(this->hostname);
@@ -1142,7 +1212,7 @@ void IrcClient::on_channel_mode(const IrcMessage& message)
     }
   this->bridge.send_message(iid, "", "Mode " + iid.get_local() +
                                       " [" + mode_arguments + "] by " + user.nick,
-                             true);
+                             true, this->is_channel_joined(iid.get_local()));
   const IrcChannel* channel = this->get_channel(iid.get_local());
   if (!channel)
     return;
@@ -1215,6 +1285,11 @@ void IrcClient::on_channel_mode(const IrcMessage& message)
     }
 }
 
+void IrcClient::set_throttle_limit(long int limit)
+{
+  this->tokens_bucket.set_limit(limit);
+}
+
 void IrcClient::on_user_mode(const IrcMessage& message)
 {
   this->bridge.send_xmpp_message(this->hostname, "",
@@ -1228,7 +1303,7 @@ void IrcClient::on_unknown_message(const IrcMessage& message)
     return ;
   std::string from = message.prefix;
   std::stringstream ss;
-  for (auto it = message.arguments.begin() + 1; it != message.arguments.end(); ++it)
+  for (auto it = std::next(message.arguments.begin()); it != message.arguments.end(); ++it)
     {
       ss << *it;
       if (it + 1 != message.arguments.end())
@@ -1252,3 +1327,94 @@ bool IrcClient::abort_on_invalid_cert() const
   return true;
 }
 #endif
+
+long int IrcClient::get_throttle_limit() const
+{
+#ifdef USE_DATABASE
+  return Database::get_irc_server_options(this->bridge.get_bare_jid(), this->hostname).col<Database::ThrottleLimit>();
+#else
+  return 10;
+#endif
+}
+
+void IrcClient::on_cap(const IrcMessage &message)
+{
+  const auto& sub_command = message.arguments[1];
+  const auto& caps = utils::split(message.arguments[2], ' ', false);
+  for (const auto& cap: caps)
+    {
+      auto it = this->capabilities.find(cap);
+      if (it == this->capabilities.end())
+        {
+          log_warning("Received a CAP message for something we didn’t ask, or that we already handled: [", cap, "]");
+          return;
+        }
+      Capability& capability = it->second;
+      if (sub_command == "ACK")
+        capability.on_ack();
+      else if (sub_command == "NACK")
+        capability.on_nack();
+      this->capabilities.erase(it);
+    }
+  if (this->capabilities.empty())
+    this->cap_end();
+}
+
+#ifdef WITH_SASL
+void IrcClient::on_authenticate(const IrcMessage &)
+{
+  if (this->sasl_state == SaslState::unneeded)
+    {
+      log_warning("Received an AUTHENTICATE command but we don’t intend to authenticate…");
+      return;
+    }
+
+  auto options = Database::get_irc_server_options(this->bridge.get_bare_jid(),
+                                                  this->get_hostname());
+  const auto auth_string = '\0' + options.col<Database::Nick>() + '\0' + options.col<Database::SaslPassword>();
+  const auto base64_auth_string = base64::encode(auth_string);
+  this->send_message({"AUTHENTICATE", {base64_auth_string}});
+}
+
+void IrcClient::on_sasl_success(const IrcMessage &)
+{
+  this->sasl_state = SaslState::success;
+  this->cap_end();
+}
+
+void IrcClient::on_sasl_failure(const IrcMessage& message)
+{
+  this->sasl_state = SaslState::failure;
+  const auto reason = message.arguments[1];
+  // Send an error message for all room that the user wanted to join
+  for (const auto& tuple: this->channels_to_join)
+  {
+    Iid iid(std::get<0>(tuple) + "%" + this->hostname, this->chantypes);
+    this->bridge.send_presence_error(iid, this->current_nick,
+                                     "cancel", "item-not-found",
+                                     "", reason);
+  }
+  this->channels_to_join.clear();
+  this->send_quit_command(reason);
+}
+
+void IrcClient::on_sasl_login(const IrcMessage &message)
+{
+  const auto& login = message.arguments[2];
+  std::string text = "Your are now logged in as " + login;
+  if (message.arguments.size() > 3)
+    text = message.arguments[3];
+  this->bridge.send_xmpp_message(this->hostname, message.prefix, text);
+}
+#endif
+
+void IrcClient::cap_end()
+{
+#ifdef WITH_SASL
+  // If we are currently authenticating through sasl, finish that before sending CAP END
+  if (this->sasl_state == SaslState::needed)
+    return;
+#endif
+  this->send_message({"CAP", {"END"}});
+  this->bridge.on_irc_client_connected(this->get_hostname());
+}
